# -*- encoding: utf-8 -*-
"""
@file
@brief Shortcut to *onnxrt*.
"""
from .onnx_inference import OnnxInference
from .onnx_micro_runtime import OnnxMicroRuntime
from .onnx_shape_inference import OnnxShapeInference
