# pylint: disable=W0611,R0401
"""
@file
@brief Shorcuts to plotting functions.
"""

from .text_plot import (
    onnx_text_plot, onnx_text_plot_tree, onnx_simple_text_plot,
    onnx_text_plot_io)
from .plotting_benchmark import plot_benchmark_metrics
from .plotting_validate_graph import plot_validate_benchmark
from .plotting_onnx import plot_onnx
