"""
@file
@brief Shortcut to *cli*.
"""
from .convert_validate import convert_validate
from .einsum import einsum_test
from .onnx_code import onnx_code
from .optimize import onnx_optim
from .validate import validate_runtime, latency
